# Copyright 2026 Infleqtion
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

import cirq

import supermarq
from supermarq.benchmarks.qaoa_vanilla_proxy import QAOAVanillaProxy


def test_qaoa_circuit() -> None:
    qaoa = QAOAVanillaProxy(4)
    assert len(qaoa.circuit().all_qubits()) == 4
    assert (
        len(
            list(qaoa.circuit().findall_operations(lambda op: isinstance(op.gate, type(cirq.CNOT))))
        )
        == 12
    )


def test_qaoa_score() -> None:
    qaoa = QAOAVanillaProxy(4)
    assert qaoa.score(supermarq.simulation.get_ideal_counts(qaoa.circuit())) > 0.99
