# Copyright 2026 Infleqtion
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from . import benchmark, converters, features, plotting, qcvv, simulation, stabilizers
from ._version import __version__
from .benchmarks import (
    bit_code,
    ghz,
    hamiltonian_simulation,
    mermin_bell,
    phase_code,
    qaoa_fermionic_swap_proxy,
    qaoa_vanilla_proxy,
    vqe_proxy,
)

__all__ = [
    "__version__",
    "benchmark",
    "bit_code",
    "converters",
    "features",
    "ghz",
    "hamiltonian_simulation",
    "mermin_bell",
    "phase_code",
    "plotting",
    "qaoa_fermionic_swap_proxy",
    "qaoa_vanilla_proxy",
    "qcvv",
    "simulation",
    "stabilizers",
    "vqe_proxy",
]
